#include <cstdlib> // defines std::size_t
#include <iterator> // defines std::iterator_traits
#include <vector> // defines std::vector
#include <list> // defines std::list
#include <functional> // defines std::less

#include <memory> // defines std::allocator


template <class T, class Container, class Compare>
class navigation_pile; // forward declaration

template <class T>
class leaf; // forward declaration

template <class T>
class branch {
public:
  branch () : p(0), parent(0) {
  };
  branch (leaf<T>* p) : p(p), parent(0) {
  };
  leaf<T>* p;
  branch<T>* parent;
  union {
    branch<T>* branch_first_child;
    leaf<T>* leaf_first_child;
  };
  union {
    branch<T>* branch_second_child;
    leaf<T>* leaf_second_child;
  };
};

template <class T>
class leaf {
public:
  leaf (const T& e) : e(e), parent(0) {
  };
  T e;
  branch<T>* parent;
};

template <class T, class Container = std::vector<T>, 
	  class Compare = std::less<typename Container::value_type> >
class navigation_pile {
 public:
  typedef typename Container::value_type value_type;
  typedef typename Container::size_type size_type;
  typedef Container container_type;
  // n: # of elements;
  // OBS OBS N present beclause of compile error
  explicit navigation_pile(size_type N,
			   const Compare& x = Compare(), 
			   const Container& Y = Container());
  template <class InputIterator>
  navigation_pile(InputIterator first, 
		  InputIterator last, 
		  const Compare& x = Compare(),
		  const Container& Y = Container() );
  ~navigation_pile();

  const  value_type& top() const;
  void push(const value_type&);
  void pop();
  bool empty() const;
  size_type size() const;
  void sort_pile(size_type, const Compare&); // Made public for testing
  
 private:
  typedef T element;
  typedef size_type level;
  typedef leaf<T>* leaf_pointer;
  typedef branch<T>* branch_pointer;
  Container A; // included for testing
  branch_pointer _root;
  leaf_pointer _last_leaf;
  Compare less;
  size_type _n;
  branch_pointer ancestor(branch_pointer, level) const;
  void dispose_pile1(branch_pointer, size_type);
  void dispose_pile2(leaf_pointer);
  void dispose_pile();
  void check_pile1(branch_pointer, size_type);
  void check_pile2(leaf_pointer);
  void check_pile();
  void  update_partial_path(leaf_pointer, branch_pointer);
  branch_pointer first_ancestor_with_two_children(leaf_pointer);
  void update_full_path(leaf_pointer);
  void update_partial_path(leaf_pointer);
  void update_partial_path(branch_pointer, branch_pointer, leaf_pointer, leaf_pointer);  
  void insert_new_element(const element&);
  void erase_last_element();
  branch_pointer bottom_up_on_special_path(branch_pointer, const element&);
  void bottom_up_update_on_special_path(branch_pointer, leaf_pointer);
  void swap_leaves(leaf_pointer, leaf_pointer);  
  size_type pow_two(size_type) const;
  size_type ceiling_log(size_type) const;
  int ilogb_zero(double) const;
};
