import benchmark
import os


class std_case(benchmark.case):
  def __init__(self, n, element, function, directory, k):
    benchmark.case.__init__(self)
    self.n = n
    self.k = k
#    self.computer = 'pc-014.diku.dk'
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
#    self.compiler = '/usr/bin/g++-3.2' 
#    self.compiler = 'g++-3' 
    self.compiler = 'g++' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/Experimental-Navigation-piles/Programs/' + directory + '/' 
      ])
    self.include_files.extend(['<vector>', '<functional>', function + '.cpp'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
         element + ',' + \
         'comparison_log<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator' +\
      '>(' + str(n) + ', ' + str(k) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.k)

#  def tidy_up(self): 
#    "Used for debugging to see the files generated"
#    pass 


class gen_case(benchmark.case):
  def __init__(self, n, element, function, directory, program, k):
    benchmark.case.__init__(self)
    self.n = n
    self.k = k
#    self.computer = 'pc-014.diku.dk'
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
#    self.compiler = '/usr/bin/g++-3.2' 
#    self.compiler = 'g++-3' 
    self.compiler = 'g++' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/Experimental-Navigation-piles/Programs/Common/',
      home + '/CPHSTL/Program/Resizable_bitarray/Implementation/',
      home + '/CPHSTL/Boost/boost_1_29_0/',
      home + '/CPHSTL/Report/Experimental-Navigation-piles/Programs/' + directory + '/' 
      ])
    self.include_files.extend(['<vector>', '<functional>', program + '.cpp', function + '.cpp'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
         element + ',' + \
         'comparison_log<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator' +\
      '>(' + str(n) + ', ' + str(k) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.k)

#  def tidy_up(self): 
#    "Used for debugging to see the files generated"
#    pass 
  

class std_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, pq_functions, algotitle, log_i, log_k, k = 0):
    benchmark.curve_suite.__init__(self)
    self.title = algotitle 
    for j in range(log_i, log_k + 1):
      self.add(std_case(1 << j, element, experiment, pq_functions, k))

class gen_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, pq_functions, algotitle, program, log_i, log_k, k = 0):
    benchmark.curve_suite.__init__(self)
    self.title = algotitle
    for j in range(log_i, log_k + 1):
      self.add(gen_case(1 << j, element, experiment, pq_functions, program, k))
