#include "binary_heap_ref.hpp"

#include "debugging.c++" // defines show

// OBS OBS N present beclause of compile error
template <class T, class Container, class Compare>
binary_heap_ref<T, Container, Compare>::binary_heap_ref(size_type N,
							const Compare& x,
							const Container& Y) {
  less_ref<enpointer, Compare> _comp = less_ref<enpointer, Compare>(x);
  for (typename Container::const_iterator i = Y.begin(); i != Y.end(); i++) {
    enpointer en = new ElementNode<T>;
    en->e = *i;
    _P.push_back(en);
  }
  make_heap(_P.begin(), _P.end(), _comp);
}

template <class T, class Container, class Compare>
template <typename InputIterator>
binary_heap_ref<T, Container, Compare>::binary_heap_ref(InputIterator first, 
						      InputIterator last,
						      const Compare& x,
							const Container& Y) {
  less_ref<enpointer, Compare> comp = less_ref<enpointer, Compare>(x);
  for (typename Container::const_iterator i = Y.begin(); i != Y.end(); i++) {
    enpointer en = new ElementNode<T>;
    en->e = *i;
    _P.push_back(en);
  }
  for (InputIterator i = first; i != last; i++) {
    enpointer en = new ElementNode<T>;
    en->e = *i;
    _P.push_back(en);
  }
  std::make_heap(_P.begin(), _P.end(), _comp);
}

template <class T, class Container, class Compare>
binary_heap_ref<T, Container, Compare>::~binary_heap_ref(){
  for (typename std::vector<enpointer>::iterator i = _P.begin(); i != _P.end(); i++) {
    enpointer b = *i;
    delete b ;
  }
}


template <class T, class Container, class Compare>
bool
binary_heap_ref<T, Container, Compare>::empty() const {
 return(_P.empty()); }


template <class T, class Container, class Compare>
typename binary_heap_ref<T, Container, Compare>::size_type
binary_heap_ref<T, Container, Compare>::size() const { 
  return(_P.size()); 
}


template <class T, class Container, class Compare>
const typename binary_heap_ref<T, Container, Compare>::value_type& 
binary_heap_ref<T, Container, Compare>::top() const {
  return(_P.front()->e);
}


template <class T, class Container, class Compare>
void 
binary_heap_ref<T, Container, Compare>::push(
const binary_heap_ref<T, Container, Compare>::value_type& x) {
    enpointer en = new ElementNode<T>;
    en->e = x;
  _P.push_back(en);
  std::push_heap(_P.begin(), _P.end(), _comp);
}


template <class T, class Container, class Compare>
void 
binary_heap_ref<T, Container, Compare>::pop() {
  enpointer i = _P.front();
  std::pop_heap(_P.begin(), _P.end(), _comp);
  _P.pop_back();
  delete i;
}

#include <algorithm> // defines std::copy
#include <is_sorted.c++> // defines cphstl::is_sorted

// Utility function used for testing
template <class T, class Container, class Compare>
void
binary_heap_ref<T, Container, Compare>::Sort_binary_heap_ref(
							   const Compare& less,
							   const Container& Y) {
  Container t(Y.size());
  std::copy(Y.begin(), Y.end(), t.begin());
  std::sort(t.begin(), t.end(), less);
  Container E(Y.size());
  // use i-- insted of int i = 0 and n++ to make is_sorted work
  int i = Y.size() - 1;
  while(binary_heap_ref<T, Container, Compare>::size() > 0) {
    E[i] = binary_heap_ref<T, Container, Compare>::top();
    binary_heap_ref<T, Container, Compare>::pop();
    i--;
  }
  assert(cphstl::is_sorted(E.begin(), E.end(), less));
  assert(std::equal(E.begin(), E.end(), t.begin()));
}
