#include <vector> // defines std::vector
#include <list> // defines std::list
#include <functional> // defines std::less

template <class T>
class ElementNode  {
public:
  ElementNode () {
  }; 
  ElementNode (const T& e) : e(e) {
  }; 
  T e;
};

template <class enpointer, class ordering>
class less_ref  {
protected:
  ordering _less;
public:
  explicit less_ref() {
  }
  explicit less_ref(const ordering& less) : _less(less){
  }
  bool operator()(const enpointer& x, const enpointer& y) const{
    return _less(x->e, y->e);
  }
  less_ref& operator=(const ordering& less) const{
    _less = less;
    return *this;
  }
};

template <class T, class Container = std::vector<T>, 
	  class Compare = std::less<typename Container::value_type> >
class binary_heap_ref {
 public:
  typedef typename Container::value_type value_type;
  typedef typename Container::size_type size_type;
  typedef Container container_type;

  // OBS OBS N present beclause of compile error
  // OBS OBS no compile error in g++ version 3.4
  // OBS OBS N also in binary_heap_ref.cpp and test.cpp
  explicit binary_heap_ref(size_type N,
			   const Compare& x = Compare(),  
			   const Container& Y = Container());
  template <class InputIterator>
  binary_heap_ref(InputIterator first, InputIterator last, 
		 const Compare& x = Compare(),
		 const Container& Y = Container());
  ~binary_heap_ref();

  bool empty() const;
  size_type size() const;
  const value_type& top() const;
  void push(const value_type& x);
  void pop();
  void Sort_binary_heap_ref(const Compare& x, const Container& Y); // Made public for testing
  private:
  typedef ElementNode<T>* enpointer;
    std::vector<enpointer> _P;
  less_ref<enpointer, Compare> _comp;
};

/*
With g++ version 3.3.5:
test.cpp: In function `int main()':
test.cpp:24: error: type specifier omitted for parameter `C'

With g++ version 3.4 okay.
*/

































