#include <vector>
#include <functional>
#include <list>

template <typename I, typename F>
class less_ref {
protected:
  F _less;

public:
  explicit less_ref (const F& less = F()) : _less(less) {
  }
  bool operator()(const I& x, const I& y) const {
    return _less(*x, *y);
  }
};

template <typename E, typename C = std::vector<E>, typename F = std::less<E> >
class pq {
public:
  typedef typename C::value_type value_type;
  typedef typename C::size_type size_type;
  typedef C container_type;

protected:
  typedef typename std::list<E>::iterator I;
  std::list<E> _element;
  std::vector<I> _pointer;
  less_ref<I, F> _less;

public:
  explicit pq(const F& = F(), const C& = C());
  bool empty() const { return _element.empty(); }
  bool size() const { return _element.size(); }
};

template <typename E, typename C, typename F>
pq<E, C, F>::pq(const F& less, const C& input) : _less(less) {
  for (typename C::const_iterator i = input.begin(); i != input.end(); ++i) {
    (*this)._element.push_back(*i);
    typename std::list<E>::iterator p = (*this)._element.end();
    --p;
    (*this)._pointer.push_back(p);
  }
}

#include <algorithm>

int main() {
  typedef int element;
  typedef std::less<element> ordering;
  typedef std::list<element>::iterator iterator;
  typedef less_ref<iterator, ordering> pointer_ordering;
  typedef std::vector<element> container;
 
  int a[7] = {1,2,3,4,5,6,7};
  container v;
  for (int i = 0; i < 7; ++i) {
    v.push_back(a[i]);
  }
  
  pq<element, container, ordering> Q(ordering(), container());

  pq<element, container, ordering> R(ordering());

  pq<element, container, ordering> S(ordering(), v);

  /*
With g++ version 3.3.5:

app-5> g++ pq.c++
pq.c++: In function `int main()':
pq.c++:66: error: type specifier omitted for parameter `v'

With g++ version 3.4 okay.
  */

  return 0;
}

  
