"""
Benchmarking various versions of heapsort

shell> echo $PYTHONPATH
/home/disk04/jyrki/CPHSTL/Tool/Benchmark/
"""

__author__ = "Jyrki Katajainen"
__email__ = "jyrki@diku.dk"
__version__ = "$Revision: 1.2 $"[11:-2]
__date__ = "$Date: 2004/01/06 16:40:23 $"[7:-2]

import benchmark
import os

class case(benchmark.case):
  def __init__(self, n, element, experiment, d, policy, heap_functions):
    benchmark.case.__init__(self)
    self.n = n
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Benchmark/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Common/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + policy + '/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + heap_functions + '/' 
      ])
    self.include_files.extend(['<vector>', '<functional>', 'heap_policy.cpp', 
      'heap_functions.cpp', experiment + '.c++'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
        str(d) + ', ' + \
         element + ',' + \
         'std::less<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator, ' +\
        'heap_policy<' +\
          str(d) + ', ' +\
          'std::vector<' + element + '>::iterator, ' +\
          'std::less<' + element + '> ' +\
        '> ' +\
      '>(' + str(n) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)

#  def tidy_up(self): 
#    "Used for debugging to see the files generated"
#    pass 
  
class curve(benchmark.curve_suite):
  def __init__(self, element, experiment, d, policy, heap_functions, log_i, log_k):
    benchmark.curve_suite.__init__(self)
    self.title = 'experiment: ' + experiment + ' source: ' + heap_functions + \
                 ' policy: ' + policy + ' arity: ' + str(d)
    for j in range(log_i, log_k + 1):
      self.add(case(1 << j, element, experiment, d, policy, heap_functions))

class plot(benchmark.plot_suite):
  def __init__(self, element, log_i, log_k):
    benchmark.plot_suite.__init__(self)
    self.title = 'Efficiency of various heap policies' 
    self.xlabel = 'n'
    self.ylabel = 'Execution time per element [in nanoseconds]'
    self.add(curve(element, 'push_pop_push', 2, 'In-place-2-policy', 'Top-down', log_i, log_k))
    self.add(curve(element, 'push_pop_push', 2, 'In-place-2-policy', 'Bottom-up', log_i, log_k))
    self.add(curve(element, 'push_pop_push', 2, 'In-place-2-policy', 'Bottom-up_move-saving', log_i, log_k))
    self.add(curve(element, 'push_pop_push', 2, 'In-place-2-policy', 'Binary-search', log_i, log_k))
    self.gnuplot_commands += """
set title '%(title)s'
set xlabel '%(xlabel)s'
set ylabel '%(ylabel)s'
set logscale x
""" % self.__dict__

if __name__ == '__main__':
  benchmark.main(
    task = sort_plot('int', 10, 11),
    runner = benchmark.gnuplot_runner
  )
