#include <iostream> // defines standard streams
#include <cassert> // defines assert
#include <cstdlib> // defines std:.size_t
#include <functional> // defines std::less
#include <algorithm> // defines std::copy
#include "input_generation.c++" // defines data generation routines

#include "heap_functions.cpp"

int main () {
  typedef std::size_t element;
  typedef element* position;
  typedef std::less<element> ordering;

  std::cout << "Testing pop_heap, make_heap and sort_heap" << std::endl;
  const std::size_t bign = 100;
  const std::size_t repetitions = 100;
  element a[bign];

  for (std::size_t n = 0; n <= bign; ++n) {

    decreasing_sequence(a, a + n);
    heapsort<2, position, ordering>(a, a + n, ordering());

    decreasing_sequence(a, a + n);
    heapsort<3, position, ordering>(a, a + n, ordering());

    decreasing_sequence(a, a + n);
    heapsort<4, position, ordering>(a, a + n, ordering());

    decreasing_sequence(a, a + n);
    heapsort<8, position, ordering>(a, a + n, ordering());

    decreasing_sequence(a, a + n);
    heapsort<16, position, ordering>(a, a + n, ordering());
  
    increasing_sequence(a, a + n);
    heapsort<2, position, ordering>(a, a + n, ordering());

    increasing_sequence(a, a + n);
    heapsort<3, position, ordering>(a, a + n, ordering());

    increasing_sequence(a, a + n);
    heapsort<4, position, ordering>(a, a + n, ordering());

    increasing_sequence(a, a + n);
    heapsort<8, position, ordering>(a, a + n, ordering());

    increasing_sequence(a, a + n);
    heapsort<16, position, ordering>(a, a + n, ordering());
  
    zero_sequence(a, a + n);
    heapsort<2, position, ordering>(a, a + n, ordering());

    zero_sequence(a, a + n);
    heapsort<3, position, ordering>(a, a + n, ordering());

    zero_sequence(a, a + n);
    heapsort<4, position, ordering>(a, a + n, ordering());

    zero_sequence(a, a + n);
    heapsort<8, position, ordering>(a, a + n, ordering());

    zero_sequence(a, a + n);
    heapsort<16, position, ordering>(a, a + n, ordering());
  
    binary_sequence(a, a + n);
    heapsort<2, position, ordering>(a, a + n, ordering());

    binary_sequence(a, a + n);
    heapsort<3, position, ordering>(a, a + n, ordering());

    binary_sequence(a, a + n);
    heapsort<4, position, ordering>(a, a + n, ordering());

    binary_sequence(a, a + n);
    heapsort<8, position, ordering>(a, a + n, ordering());

    binary_sequence(a, a + n);
    heapsort<16, position, ordering>(a, a + n, ordering());

    for (std::size_t i = 0; i < repetitions; ++i) {
      random_sequence(a, a + n);
      heapsort<2, position, ordering>(a, a + n, ordering());

      random_sequence(a, a + n);
      heapsort<3, position, ordering>(a, a + n, ordering());

      random_sequence(a, a + n);
      heapsort<4, position, ordering>(a, a + n, ordering());

      random_sequence(a, a + n);
      heapsort<8, position, ordering>(a, a + n, ordering());

      random_sequence(a, a + n);
      heapsort<16, position, ordering>(a, a + n, ordering());
    }
    std::cout << n << " ok" << std::endl;
  }

  std::cout << "Testing push_heap" << std::endl;

  typedef void routine(position, position);
  routine* generate_sequence[] = {
    increasing_sequence,
    decreasing_sequence,
    zero_sequence,
    random_sequence,
    binary_sequence
  };

  for (std::size_t n = 1; n <= bign; ++n) {
    for (unsigned int i = 0; i < sizeof(generate_sequence) / sizeof(routine*); ++i) {
      (void) generate_sequence[i](a, a + n);
      element b[bign];
      std::copy(a, a + n, b);
      typedef heap_policy<8, position, ordering> policy;
      policy p(0);
      make_heap<8, position, ordering, policy>(a, ordering(), p);
      std::make_heap(b, b, ordering());
      for (unsigned int k = 1; k <= n; ++k) {
        push_heap<8, position, ordering, policy>(a, ordering(), p);
	bool result;
	result = is_heap<8, position, ordering, policy>(a, a + p.size(), ordering(), p);
        assert(result);
      }
    }
    std::cout << n << " ok" << std::endl;
  }

  return 0;
}
