#ifndef __UTILITY_FUNCTIONS__
#define __UTILITY_FUNCTIONS__

#include <iostream>
#include <cassert>

template <int d, typename position, typename level, typename index, typename element, 
  typename ordering, typename policy>
index
correct_place(position a, index l, level h, element x, const ordering& less, policy& p) {
  
  /* exp. part: parameters needed */
  
  if (h == 0) return l;
  
  level delta = 1;
  index i;
  index j = l;
  while (true) {
    if (delta > h) {
      i = p.ancestor(j, h, h);
      break;
    }
    i = p.ancestor(j, h, delta);
    if (less(a[i], x)) {
      h = h - delta;
      delta = 2 * (delta + 1) - 1;
      j = i;
    }
    else {
      h = delta;
      break;
    }
  }
  
  /* bin. part: i, j, and h needed */
        
  index k = j;
  while (h > 1) {
    level g = h / 2;
    j = p.ancestor(k, h, g);
    if (less(a[j], x)) {
      h = h - g;
      k = j;
    }
    else {
      h = g;
      i = j;
    }
  }
  
  if (less(x, a[i])) {
    return k;
  }
  else {
    return i;
  }
}

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
void
sift_down(position a, index j, element x, const ordering& less, policy& p) {
  // std::cout << "before sift_down j, x: " << j << " " << x << std::endl;
  // show(a, a + p.size());
  
  typedef typename policy::level level; 
  level h = 0;
  index k;
  index i;
  
  const index last_full_branch = p.parent(p.last_leaf() + 1) - 1;
  
  while (j <= last_full_branch) {
    k = p.top_all_present(a, j, less);
    p.update(a, j, a[k]);
    j = k;
    h = h + 1;
  }
  
  k = p.first_child(j);
  if (k < p.size()) {
    k = p.top_some_absent(a, j, less);      
    p.update(a, j, a[k]);
    j = k;
    h = h + 1;
  }
  
  i = correct_place<d, position, level, index, element, ordering, policy>(a, j, h, x, less, p);
  
  while (i != j) {
    p.update(a, j, a[p.parent(j)]);
    j = p.parent(j);
  }
  
  p.update(a, i, x);
  
  //std::cout << "after sift_down: " << std::endl;
  //show(a, a + p.size());
} 

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
void
sift_up(position a, index j, element x, const ordering& less, policy& p) {
  //std::cout << "before sift_up j: " << j << std::endl;
  //show(a, a + p.size());
  
  typedef typename policy::level level; 
  level h;
  index i;
  
  h = p.depth(j);
  i = correct_place<d, position, level, index, element, ordering, policy>(a, j, h, x, less, p);
  
  while (i != j) {
    p.update(a, j, a[p.parent(j)]);
    j = p.parent(j);
  }
  
  p.update(a, i, x);
  
  //std::cout << "after sift_up" << std::endl;
  //show(a, a + p.size());
} 

#endif // __UTILITY_FUNCTIONS__


