#include <iostream>
#include <functional>
#include <cassert>

#include "heap_policy.cpp" // defines policy
#include "utility_functions.cpp" // defines utility functions called by heap functions
#include "heap_functions.cpp" // defines make_heap

int main() {

  int n = 17;
  typedef int element;
  typedef element* position;
  typedef std::less<element> ordering;
  typedef heap_policy<8, position, ordering> policy;
  ordering less = ordering();
  policy p(n);
  int a[17] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16};

  std::cout <<
    "testing is_root " << std::endl;
  assert(p.is_root(0) == true);
  assert(p.is_root(1) == false);
  assert(p.is_root(2) == false);
  assert(p.is_root(3) == false);
  assert(p.is_root(4) == false);
  assert(p.is_root(5) == false);

  std::cout <<
    "testing is_first_child " << std::endl;
  assert(p.is_first_child(0) == false);
  assert(p.is_first_child(1) == true);
  assert(p.is_first_child(2) == false);
  assert(p.is_first_child(3) == false);
  assert(p.is_first_child(4) == false);
  assert(p.is_first_child(5) == false);
  assert(p.is_first_child(6) == false);
  assert(p.is_first_child(7) == false);
  assert(p.is_first_child(8) == false);
  assert(p.is_first_child(9) == true);
  assert(p.is_first_child(10) == false);


  std::cout <<
    "testing root " << std::endl;
  assert(p.root() == 0);

  std::cout <<
    "testing size " << std::endl;
  assert(p.size() == n);

  std::cout <<
    "testing depth " << std::endl;
  assert(p.depth(0) == 0);
  assert(p.depth(1) == 1);
  assert(p.depth(2) == 1);
  assert(p.depth(3) == 1);
  assert(p.depth(4) == 1);
  assert(p.depth(5) == 1);
  assert(p.depth(6) == 1);
  assert(p.depth(7) == 1);
  assert(p.depth(8) == 1);
  assert(p.depth(9) == 2);
  assert(p.depth(10) == 2);

  std::cout <<
    "testing leftmost_node "  << std::endl;
  assert(p.leftmost_node(2) == 9);

  std::cout <<
    "testing first_child "  << std::endl;
  assert(p.first_child(0) == 1);
  assert(p.first_child(1) == 9);
  assert(p.first_child(2) == 17);
  assert(p.first_child(3) == 25);

  std::cout <<
    "testing parent " << std::endl;
  assert(p.parent(1) == 0);
  assert(p.parent(2) == 0);
  assert(p.parent(3) == 0);
  assert(p.parent(4) == 0);
  assert(p.parent(9) == 1);
  assert(p.parent(10) == 1);
  assert(p.parent(17) == 2);
  assert(p.parent(19) == 2);
  assert(p.parent(25) == 3);
  assert(p.parent(28) == 3);
  assert(p.parent(33) == 4);

  std::cout <<
    "testing ancestor " << std::endl;
  assert(p.ancestor(0, 0, 0) == 0);
  assert(p.ancestor(1, 1, 0) == 1);
  assert(p.ancestor(1, 1, 1) == 0);
  assert(p.ancestor(2, 1, 1) == 0);
  assert(p.ancestor(3, 1, 0) == 3);
  assert(p.ancestor(3, 1, 1) == 0);
  assert(p.ancestor(9, 2, 0) == 9);
  assert(p.ancestor(9, 2, 1) == 1);
  assert(p.ancestor(9, 2, 2) == 0);
  assert(p.ancestor(42, 2, 0) == 42);
  assert(p.ancestor(42, 2, 1) == 5);
  assert(p.ancestor(42, 2, 2) == 0);

  std::cout << 
    "testing top_all_present " << std::endl;
  assert(p.top_all_present(a, 0, less) == 8);

  std::cout << 
    "testing top_all_present " << std::endl;
  assert(p.top_all_present(a, 1, less) == 16);


  std::cout << 
    "testing erase_last_leaf " << std::endl;
  p.erase_last_leaf();
  assert(p.size() == n - 1);
	
  std::cout << 
    "testing insert_new_leaf " << std::endl;
  p.insert_new_leaf();
  assert(p.size() == n);
  p.insert_new_leaf();
  assert(p.leftmost_node(p.depth(p.last_leaf())) == 9);


  std::cout << 
    "testing top_some_absent " << std::endl;
  p.erase_last_leaf();
  p.erase_last_leaf();
  p.erase_last_leaf();
  p.erase_last_leaf();

  assert(p.top_some_absent(a, 1, less) == 13);

  p.erase_last_leaf();
  assert(p.top_some_absent(a, 1, less) == 12);

  p.erase_last_leaf();
  assert(p.top_some_absent(a, 1, less) == 11);


  return 0;
}
