#ifndef __HEAP_POLICY__
#define __HEAP_POLICY__

#include "heap_policy.hpp" // declares heap_policy
#include <cmath> // defines std::ilogb

extern int ilogb(double) throw();

#ifdef BUFFER
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::heap_policy (
  heap_policy<d, position, ordering>::index n
): n(n) {
  if (n <= index(d) + 1) { l = 1;}
  else {
    l = d * ((n - 1) / index(d) - 1) + 1;
    if ( l + index(d) != n) {
      l = l + index(d);
    }
  }
}
#else
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::heap_policy (
  heap_policy<d, position, ordering>::index n
): n(n) {
}
#endif

template <int d, typename position, typename ordering>
bool
heap_policy<d, position, ordering>::is_root (
  heap_policy<d, position, ordering>::index i
) const {
  return (i == 0);
}

template <int d, typename position, typename ordering>
bool
heap_policy<d, position, ordering>::is_first_child (
  heap_policy<d, position, ordering>::index i
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  return (i - (i / index(d)) * index(d) == 1);
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::size () const {
  return n;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::level
heap_policy<d, position, ordering>::depth (
  heap_policy<d, position, ordering>::index i
) const {
  typedef typename heap_policy<d, position, ordering>::level level;
  typedef typename heap_policy<d, position, ordering>::index index;
  level h = 0;
  index m = 1;
  index nr = 0;
  while (i > nr) {
    m = m * index(d);
    nr = nr + m;
    ++h;
  }
  return h;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::root () const {
  return 0;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::leftmost_node (
  heap_policy<d, position, ordering>::level h
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  index r = 1;
  while (h > 0) {
    r = r * index(d);
    --h;
  }
  r = (r - 1) / index(d - 1);
  return r;
}

#ifdef BUFFER
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::last_leaf () const {
  return l - 1;
}
#else
template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
    heap_policy<d, position, ordering>::last_leaf () const {
  return n - 1;
}
#endif

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::first_child (
  heap_policy<d, position, ordering>::index p
) const {
  return d * p + 1;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::parent (
  heap_policy<d, position, ordering>::index c
) const {
  return (c - 1) / d;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::ancestor (
  heap_policy<d, position, ordering>::index j,
  heap_policy<d, position, ordering>::level h,
  heap_policy<d, position, ordering>::level Delta
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  index t = 1;
  while (Delta > 0) {
    t = t * index(d);
    --Delta;
  }
  index i = (t - 1) / index(d - 1);
  return (j - i) / t;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::top_some_absent (
  position a,
  heap_policy<d, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  typedef typename heap_policy<d, position, ordering>::element element;
  index j = heap_policy<d, position, ordering>::first_child(i);
  index k = (j + index(d) < n) ? j + index(d) : n;
  index maxc = j;
  element max = a[j];
  ++j;
  while (j < k) {
    if (less(max, a[j])) {
      maxc = j;
      max = a[j];
    }
    ++j;
  }
  return maxc;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::top_all_present (
  position a,
  heap_policy<d, position, ordering>::index i,
  const ordering& less
) const {
  return heap_policy<d, position, ordering>::top_some_absent(a, i, less);
}

template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::update (
  position a,
  heap_policy<d, position, ordering>::index i,
  const heap_policy<d, position, ordering>::element& x
) {
  a[i] = x;
}

#ifdef BUFFER
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::erase_last_leaf (
) {
  --n;
  if (l == n) {
    if ( n <= index(d) ) {l = 1;}
    else {l = l - index(d);}
  }
}
#else
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::erase_last_leaf (
) {
  --n;
}
#endif

#ifdef BUFFER
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::insert_new_leaf (
) {
  ++n;
  if (( l + 2 * index(d)) == n ) {
    l = l + index(d);
  }
}
#else
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::insert_new_leaf (
) {
  ++n;
}
#endif

#endif // __HEAP_POLICY__
