#include <iostream> // defines standard streams
#include <functional> // defines std::less
#include <cstdlib>  // defines std::size_t
#include <cmath> // defines std::pow
#include <vector> // defines std::vector

#include "helperfunctions.cpp"
#include "bigint.c++"

const bigint zero = bigint((unsigned int)(0));
const bigint large_number = bigint((unsigned int)(1 << 20));

// larger to smaller
template <typename element, typename ordering>
class sequential_write_1 {
public:
  // n must be a power of 2!
  sequential_write_1(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i < n; ++i) {
      a[i] = zero;
    }
    x = large_number;
  }
  void operator()() {
    unsigned int i;
    unsigned int mask = a.size() - 1;
    unsigned int step = 1;
    for (i = step; i != 0; i = (i + step) & mask) {
      a[i] = x;
    }
  }
private:
  std::vector<element> a;
  element x;
};

// smaller to larger
template <typename element, typename ordering>
class sequential_write_2 {
public:
  // n must be a power of 2!
  sequential_write_2(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i < n; ++i) {
      a[i] = large_number;
    }
    x = zero;
  }
  void operator()() {
    unsigned int i;
    unsigned int mask = a.size() - 1;
    unsigned int step = 1;
    for (i = step; i != 0; i = (i + step) & mask) {
      a[i] = x;
    }
  }
private:
  std::vector<element> a;
  element x;
};

// larger to smaller
template <typename element, typename ordering>
class arbitrary_write_1 {
public:
  // n must be a power of 2!
  arbitrary_write_1(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i < n; ++i) {
      a[i] = zero;
    }
    x = large_number;
  }
  void operator()() {
    unsigned int i;
    unsigned int mask = a.size() - 1;
    unsigned int step = 617;
    for (i = step; i != 0; i = (i + step) & mask) {
      a[i] = x;
    }
  }
private:
  std::vector<element> a;
  element x;
};

// smaller to larger
template <typename element, typename ordering>
class arbitrary_write_2 {
public:
  // n must be a power of 2!
  arbitrary_write_2(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i < n; ++i) {
      a[i] = large_number;
    }
    x = zero;
  }
  void operator()() {
    unsigned int i;
    unsigned int mask = a.size() - 1;
    unsigned int step = 617;
    for (i = step; i != 0; i = (i + step) & mask) {
      a[i] = x;
    }
  }
private:
  std::vector<element> a;
  element x;
};

// larger to smaller
template <typename element, typename ordering>
class sequential_read_1 {
public:
  // n must be a power of 2!
  sequential_read_1(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i < n; ++i) {
      a[i] = large_number;
    }
    x = zero;
  }
  void operator()() {
    unsigned int i;
    unsigned int mask = a.size() - 1;
    unsigned int step = 1;
    for (i = step; i != 0; i = (i + step) & mask) {
      x = a[i];
    }
  }
private:
  std::vector<element> a;
  element x;
};

// smaller to larger
template <typename element, typename ordering>
class sequential_read_2 {
public:
  // n must be a power of 2!
  sequential_read_2(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i < n; ++i) {
      a[i] = zero;
    }
    x = large_number;
  }
  void operator()() {
    unsigned int i;
    unsigned int mask = a.size() - 1;
    unsigned int step = 1;
    for (i = step; i != 0; i = (i + step) & mask) {
      x = a[i];
    }
  }
private:
  std::vector<element> a;
  element x;
};

// larger to smaller
template <typename element, typename ordering>
class arbitrary_read_1 {
public:
  // n must be a power of 2!
  arbitrary_read_1(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i < n; ++i) {
      a[i] = large_number;
    }
    x = zero;
  }
  void operator()() {
    unsigned int i;
    unsigned int mask = a.size() - 1;
    unsigned int step = 617;
    for (i = step; i != 0; i = (i + step) & mask) {
      x = a[i];
    }
  }
private:
  std::vector<element> a;
  element x;
};

// smaller to larger
template <typename element, typename ordering>
class arbitrary_read_2 {
public:
  // n must be a power of 2!
  arbitrary_read_2(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i < n; ++i) {
      a[i] = zero;
    }
    x = large_number;
  }
  void operator()() {
    unsigned int i;
    unsigned int mask = a.size() - 1;
    unsigned int step = 617;
    for (i = step; i != 0; i = (i + step) & mask) {
      x = a[i];
    }
  }
private:
  std::vector<element> a;
  element x;
};

// smaller to larger
template <typename element, typename ordering>
class comparison_1 {
public:
  // n must be a power of 2!
  comparison_1(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i < n; ++i) {
      a[i] = zero;
    }
    x = large_number;
    less = ordering();
  }
  void operator()() {
    unsigned int i;
    unsigned int mask = a.size() - 1;
    unsigned int step = 1;
    for (i = step; i != 0; i = (i + step) & mask) {
      result = less(a[i], x);
    }
  }
private:
  std::vector<element> a;
  element x;
  volatile bool result;
  ordering less;
};

// two equal large
template <typename element, typename ordering>
class comparison_2 {
public:
  // n must be a power of 2!
  comparison_2(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i < n; ++i) {
      a[i] = large_number;
    }
    x = large_number;
    less = ordering();
  }
  void operator()() {
    unsigned int i;
    unsigned int mask = a.size() - 1;
    unsigned int step = 1;
    for (i = step; i != 0; i = (i + step) & mask) {
      result = less(a[i], x);
    }
  }
private:
  std::vector<element> a;
  element x;
  volatile bool result;
  ordering less;
};

int main() {
  typedef bigint element;
  typedef std::less<element> ordering;
  typedef sequential_write_1<element, ordering> seq_write_functor_1;  
  typedef sequential_write_2<element, ordering> seq_write_functor_2;  
  typedef arbitrary_write_1<element, ordering> arb_write_functor_1;  
  typedef arbitrary_write_2<element, ordering> arb_write_functor_2;  
  typedef sequential_read_1<element, ordering> seq_read_functor_1;  
  typedef sequential_read_2<element, ordering> seq_read_functor_2;  
  typedef arbitrary_read_1<element, ordering> arb_read_functor_1;  
  typedef arbitrary_read_2<element, ordering> arb_read_functor_2;  
  typedef comparison_1<element, ordering> comp_functor_1;  
  typedef comparison_2<element, ordering> comp_functor_2;

  const double unit = std::pow(double(10), 9);
  const std::size_t min_log_n = 10;
  const std::size_t max_log_n = 11;

/*
  { 
    std::cout << "cost of a sequential write (larger to smaller)" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(seq_write_functor_1(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }

  { 
    std::cout << "cost of a sequential write (smaller to larger)" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(seq_write_functor_2(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }

  { 
    std::cout << "cost of an arbitrary write (larger to smaller)" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(arb_write_functor_1(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }

  { 
    std::cout << "cost of an arbitrary write (smaller to larger)" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(arb_write_functor_2(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }

  { 
    std::cout << "cost of a sequential read (larger to smaller)" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(seq_read_functor_1(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }

  { 
    std::cout << "cost of a sequential read (smaller to larger)" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(seq_read_functor_2(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }
*/
  { 
    std::cout << "cost of an arbitrary read (larger to smaller)" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(arb_read_functor_1(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }
 
  { 
    std::cout << "cost of an arbitrary read (smaller to larger)" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(arb_read_functor_2(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }

  { 
    std::cout << "cost of a comparison (larger to smaller)" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(comp_functor_1(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }

  { 
    std::cout << "cost of a comparison (two equal large)" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(comp_functor_2(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }
  return 0;
}
