#include <iostream> // defines standard streams
#include <functional> // defines std::less
#include <cstdlib>  // defines std::size_t
#include <cmath> // defines std::pow
#include <vector> // defines std::vector
#include <cassert> // defines assert

#include "../Common/helperfunctions.cpp"

typedef unsigned int element;

template <typename element>
class division {
public:
  division(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i != n; ++i) {
      a[i] = i & (1 << 7 - 1);
    }
  }
  void operator()() {
    for (unsigned int i = 0; i != a.size(); ++i) {
      a[i] = a[i] / 3;
    }
  }
private:
  std::vector<element> a;
};

static element div3[] = {
  0,0,0,1,1,1,2,2,2,3,3,3,4,4,4,5,5,5,6,6,6,7,7,7,8,8,8,9,9,9,
  10,10,10,11,11,11,12,12,12,13,13,13,14,14,14,15,15,15,16,16,16,17,17,17,18,18,18,19,19,19,
  20,20,20,21,21,21,22,22,22,23,23,23,24,24,24,25,25,25,26,26,26,27,27,27,28,28,28,29,29,29,
  30,30,30,31,31,31,32,32,32,33,33,33,34,34,34,35,35,35,36,36,36,37,37,37,38,38,38,38,38,39,39,39,
  40,40,40,41,41,41,42,42,42
};
template <typename element>
class div3_table {
public:
  div3_table(unsigned int n) {
    a.resize(n);
    for (unsigned int i = 0; i != n; ++i) {
      a[i] = i & (1 << 7 - 1);
    }
  }
  void operator()() {
    for (unsigned int i = 0; i != a.size(); ++i) {
      a[i] = div3[a[i]];
    }
  }
private:
  std::vector<element> a;
};

int main() {
  const double unit = std::pow(double(10), 9);
  const std::size_t min_log_n = 10;
  const std::size_t max_log_n = 24;

  { 
    std::cout << "cost of a[i] = a[i] / 3" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(division<element>(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }

  { 
    std::cout << "cost of a[i] = div3[a[i]]" << std::endl;
    for (std::size_t i = min_log_n; i <= max_log_n; ++i) {
      const std::size_t n = 1 << i;
      double time = measure_same_object(div3_table<element>(n));
      std::cout  << n << " " << unit * time / double(n) << std::endl;
    }
    std::cout << std::endl;
  }

  return 0;
}








