"""
Benchmarking of local versus global moves

"""
import benchmark
import os

class if_case(benchmark.case):
  def __init__(self, n, element, function, directory, include_file):
    benchmark.case.__init__(self)
    self.n = n
#    self.computer = 'pc-014.diku.dk'
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
#    self.compiler = '/usr/bin/g++-3.2' 
#    self.compiler = '/usr/bin/g++-3.2.2' 
    self.compiler = '/usr/bin/g++' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Microbenchmarks/' + directory + '/' 
      ])
    self.include_files.extend(['<vector>', '<functional>', include_file + '.c++', function + '.c++'])
    self.dual_exists = 0
    self.constructor_call = \
      'experiment<' + \
         element + ',' + \
         'std::less<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator' +\
      '>(' + str(n) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)
#  def tidy_up(self): 
#    "Used for debugging to see the files generated"
#    pass 

class if_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, functions, include_file, log_i, log_k):
    benchmark.curve_suite.__init__(self)
    self.title = 'experiment: ' + experiment + ' source: ' + functions 
    for j in range(log_i, log_k + 1):
      self.add(if_case(1 << j, element, experiment, functions, include_file))

class if_plot(benchmark.plot_suite):
  def __init__(self, element, log_i, log_k):
    benchmark.plot_suite.__init__(self)
    self.title = 'Efficiency of local moves versus global moves for random integers' 
    self.xlabel = 'n'
    self.ylabel = 'Execution time per element [in nanoseconds]'
    self.add(if_curve(element, 'cost_depth', 'DepthBinary', 'depth', log_i, log_k))
    self.add(if_curve(element, 'cost_depth', 'DepthIf', 'depth', log_i, log_k))
    self.add(if_curve(element, 'cost_depth', 'DepthWhile', 'depth', log_i, log_k))
    self.gnuplot_commands += """
set title '%(title)s'
set xlabel '%(xlabel)s'
set ylabel '%(ylabel)s'
set logscale x
""" % self.__dict__

if __name__ == '__main__':
  benchmark.main(
    task = if_plot('int', 10, 25),
    runner = benchmark.gnuplot_runner
  )
