#ifndef __UTILITY_FUNCTIONS__
#define __UTILITY_FUNCTIONS__

#include <iostream>

template <int d, typename position, typename index, typename element,
  typename ordering, typename policy>
void
sift_down(position a, index i, element x, const ordering& less, policy& p) {
  //std::cout << "before sift_down i,x: " << i << " " << x << std::endl;
  //show(a, a + p.size());

  index j = p.first_child(i);
  const index last_full_branch = p.parent(p.last_leaf() + 1) - 1;
  
  while (i <= last_full_branch) {
    j = p.top_all_present(a, i, less);
    if (less(x, a[j])) {
      p.update(a, i, a[j]);
    }
    else {
      p.update(a, i, x);
      return;
    }
    i = j;
  }
  
  j = p.first_child(i);
  if (j >= p.size()) {
    p.update(a, i, x);
    return;
  };
  j = p.top_some_absent(a, i, less);
  if (less(x, a[j])) {
    p.update(a, i, a[j]);
    p.update(a, j, x);
  }
  else {
    p.update(a, i, x);
  }

 //std::cout << "after sift_down: " << std::endl;
 //show(a, a + p.size());
}


#include <vector> // defines std::vector

template <int d, typename position, typename index, typename element,
  typename ordering, typename policy>
void
sift_up_buffer(position a, const ordering& less, policy& p) {
  //std::cout << "before sift_up_buffer" << std::endl;
  //show(a, a + p.size());

/*
  // rp Root Patch
  vector<index> rp;

  index k = p.size() - 1 - d;
  rp.push_back(k);
  // Build Patch to Root
  do {
    k = p.parent(k);
    rp.push_back(k);
  } while (k != p.root());
  std::reverse(rp.begin(), rp.end()); //reverse_iterator do not work ???

  index buffer_parent = p.parent(p.last_leaf() + 1);
  element tmp_elem;
  index j = p.top_all_present(a, buffer_parent, less);   

  for (vector<index>::const_iterator i = rp.begin(); i != rp.end(); ++i) {
    if (less(a[*i], a[j])) {
      tmp_elem = a[*i];
      a[*i] = a[j];
      p.update(a, j, tmp_elem);
      j = p.top_all_present(a, buffer_parent, less);   
    }
  }
} 
*/

  typedef typename policy::level level;
  index i;
  index l = p.last_leaf() + 1;
  index buffer_parent = p.parent(l);
  level h = p.depth(l);
  level hl = p.depth(l);
  element tmp_elem;
  index j = p.top_all_present(a, buffer_parent, less);   

  do {
    i = p.ancestor(l, hl, h);
    if (less(a[i], a[j])) {
      tmp_elem = a[i];
      a[i] = a[j];
      p.update(a, j, tmp_elem);
      j = p.top_all_present(a, buffer_parent, less);   
    }
    h--;
  } while (h >= 0);

  //std::cout << "after sift_up_buffer" << std::endl;
  //show(a, a + p.size());
 }

#endif // __UTILITY_FUNCTIONS__
