"""
Benchmarking the partial_sort() function in the header algorithm

shell> echo $PYTHONPATH
/home/disk04/jyrki/CPHSTL/Tool/Benchmark/
"""

__author__ = "Jyrki Katajainen"
__email__ = "jyrki@diku.dk"
__version__ = "$Revision: 1.5 $"[11:-2]
__date__ = "$Date: 2003/06/08 15:43:15 $"[7:-2]

import benchmark
import os

class partial_sort_case(benchmark.case):
  def __init__(self, k, n, element, namespace, header):
    benchmark.case.__init__(self)
    self.k = k
    self.n = n
    self.computer = 'pc-014.diku.dk'
    self.compiler = '/scratch/g++/bin/g++'
    self.compiler_options.extend(['-O3', '-DNAMESPACE=' + namespace])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Tool/Benchmark/Examplebase/', 
      home + '/CPHSTL/Program/Algorithm/', 
      home + '/CPHSTL/Program/Optimization/'])
    self.include_files.extend([header, 'partial_sort_benchmark.c++'])
    self.dual_exists = 1
    self.constructor_call = 'experiment<' + element + '>(' + str(k) + ',' + str(n) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (float(self.k * 100) / self.n, float(self.driver_output) / self.n)

class partial_sort_curve(benchmark.curve_suite):
  def __init__(self, n, element, namespace, header):
    benchmark.curve_suite.__init__(self)
    self.title = namespace + '::partial_sort()' 
    for k in range(0, n + 1, n / 10):
      self.add(partial_sort_case(k, n, element, namespace, header))

class partial_sort_plot(benchmark.plot_suite):
  """
  Measure the efficiency of the partial_sort() function for random integers
  """
  def __init__(self, n):
    benchmark.plot_suite.__init__(self)
    self.title = 'Efficiency of the partial_sort() function for ' + str(n) + ' random integers' 
    self.xlabel = 'Sorted portion [in procents]'
    self.ylabel = 'Execution time per element [in nanoseconds]'
    self.add(partial_sort_curve(n, 'int', 'std', '<algorithm>'))
    self.add(partial_sort_curve(n, 'int', 'cphstl', 'algorithm')) 
    self.gnuplot_commands += """
set title '%(title)s'
set xlabel '%(xlabel)s'
set ylabel '%(ylabel)s'
""" % self.__dict__

class curve_collection(benchmark.suite):
  def __init__(self):
    benchmark.suite.__init__(self)
    for i in range(10, 21):
      self.add(partial_sort_plot(1 << i))

if __name__ == '__main__':
  benchmark.main(task = curve_collection(), runner = benchmark.gnuplot_runner)

