"""
Benchmarking the execution time of an empty loop

shell> echo $PYTHONPATH
/home/disk04/jyrki/CPHSTL/Tool/Benchmark/
"""

__author__ = "Jyrki Katajainen"
__email__ = "jyrki@diku.dk"
__version__ = "$Revision: 1.6 $"[11:-2]
__date__ = "$Date: 2003/06/08 15:43:15 $"[7:-2]

import benchmark
import os

class spin_case(benchmark.case):
  def __init__(self, n):
    benchmark.case.__init__(self)
    self.n = n
    self.computer = 'pc-014.diku.dk'
    self.compiler = '/scratch/g++/bin/g++'
    self.compiler_options.extend(['-O3'])
    self.include_files.extend(['spin_benchmark.c++'])
    self.dual_exists = 0
    self.constructor_call = 'spin(' + str(n) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)

class spin_curve(benchmark.curve_suite):
  def __init__(self, low, high):
    benchmark.curve_suite.__init__(self)
    self.title = 'i = n; while (--i >= 0);' 
    for k in range(low, high + 1):
      self.add(spin_case(k))

class spin_plot(benchmark.plot_suite):
  def __init__(self, low, high):
    benchmark.plot_suite.__init__(self)
    self.title = 'Execution time of an empty loop' 
    self.xlabel = 'n'
    self.ylabel = 'Execution time per iteration [in nanoseconds]'
    self.add(spin_curve(low, high))
    self.gnuplot_commands += """
set title '%(title)s'
set xlabel '%(xlabel)s'
set ylabel '%(ylabel)s'
""" % self.__dict__

if __name__ == '__main__':
  benchmark.main(task = spin_plot(2, 10), runner = benchmark.gnuplot_runner)

